@@ -1,60 +1,20 @@
 r"""
-======================
-Hamiltonian drive term
-======================
+==========================================
+Drive Hamiltonian (:mod:`qudit_sim.drive`)
+==========================================
 
-The drive Hamiltonian for a given channel :math:`j`, qudit :math:`k`, level :math:`l` is
+.. currentmodule:: qudit_sim.drive
 
-.. math::
-
-    \tilde{H}_{\mathrm{d}}|_{jk}^{l} = \alpha_{jk} \frac{\Omega_j}{2} \left( r_j(t) e^{-i (\nu_j t - \rho_{jk})} + \mathrm{c.c.} \right) \left( e^{i (\xi_k^{l} t + \phi_k^{l})} \sqrt{l+1} |l + 1 \rangle_k \langle l |_k  + \mathrm{h.c.} \right).
-
-Let
-
-.. math::
-
-    R_{jk}(t) = \alpha_{jk} e^{i\rho_{jk}} \frac{\Omega_j}{2} r_j(t)
-
-and
-
-.. math::
-
-    A^{l}_{k} & = e^{-i \phi^{l}_{k}} \sqrt{l + 1} | l \rangle_k \langle l + 1 |_k, \\
-    X^{l}_{k} & = A^{l\dagger}_{k} + A^{l}_{k} \\
-    Y^{l}_{k} & = i(A^{l\dagger}_{k} - A^{l}_{k}).
-
-Then the drive term above is
-
-.. math::
-
-    \tilde{H}_{\mathrm{d}}|_{jk}^{l} = & R_{jk}(t) e^{-i (\nu_j - \xi_k^{l}) t} A^{l\dagger}_{k} + R^{*}_{jk}(t) e^{i (\nu_j - \xi_k^{l}) t} A^{l}_{k}
-                                       + R^{*}_{jk}(t) e^{i (\nu_j + \xi_k^{l}) t} A^{l\dagger}_{k} + R_{jk}(t) e^{-i (\nu_j + \xi_k^{l}) t} A^{l}_{k} \\
-                                     = & \mathrm{Re}[R_{jk}(t) e^{-i (\nu_j - \xi_k^{l}) t} + R_{jk}(t) e^{-i (\nu_j + \xi_k^{l}) t}] X^{l}_{k}
-                                       + \mathrm{Im}[R_{jk}(t) e^{-i (\nu_j - \xi_k^{l}) t} - R_{jk}(t) e^{-i (\nu_j + \xi_k^{l}) t}] Y^{l}_{k} \\
-                                     = & 2 \mathrm{Re}[R_{jk}(t) e^{-i \nu_j t}] [\cos(\xi_k^{l} t) X^{l}_{k} + \sin(\xi_k^{l} t) Y^{l}_{k}].
-
-With the rotating-wave approximation we instead have
-
-.. math::
-
-    \bar{H}_{\mathrm{d}}|_{jk}^{l} = & R_{jk}(t) e^{-i (\nu_j - \xi_k^{l}) t} A^{l\dagger}_{k} + R^{*}_{jk}(t) e^{i (\nu_j - \xi_k^{l}) t} A^{l}_{k} \\
-                                     = & \mathrm{Re}[R_{jk}(t) e^{-i (\nu_j - \xi_k^{l}) t}] X^{l}_{k} + \mathrm{Im}[R_{jk}(t) e^{-i (\nu_j - \xi_k^{l}) t}] Y^{l}_{k}.
-
-The representation in terms of :math:`X^{l}_{k}` and :math:`Y^{l}_{k}` operators has several advantages over using :math:`A^{l}_{k}` and :math:`A^{l\dagger}_{k}`:
-
-- When :math:`r_j(t)` is a callable, QuTiP `sesolve` seems to run slightly faster when :math:`X` and :math:`Y` with real coefficients are passed as Hamiltonian terms.
-- The Hamiltonian is manifestly Hermitian.
-- For a pure real or imaginary :math:`R_{jk}(t)`, on-resonant (:math:`\nu_j = \xi_k^{l}`) drive, the RWA Hamiltonian reduces to a single term.
+See :ref:`drive-hamiltonian` for theoretical background.
 """
 
 from typing import Callable, Optional, Union, Tuple
 from dataclasses import dataclass
 import numpy as np
 
-from .pulse import PulseSequence
-from .util import HamiltonianCoefficient
-
-REL_FREQUENCY_EPSILON = 1.e-7
+# Type for the callable time-dependent Hamiltonian coefficient
+CallableCoefficient = Callable[[Union[float, np.ndarray], dict], Union[complex, np.ndarray]]
+HamiltonianCoefficient = Union[str, np.ndarray, CallableCoefficient]
 
 def cos_freq(freq, phase=0.):
     """`cos(freq * t + phase)`"""
@@ -117,7 +77,7 @@ class DriveTerm:
             a constant phase. None otherwise.
     """
     frequency: Optional[float] = None
-    amplitude: Union[float, complex, str, np.ndarray, PulseSequence, Callable] = 1.+0.j
+    amplitude: Union[float, complex, str, np.ndarray, Callable] = 1.+0.j
     constant_phase: Optional[float] = None
 
     def generate_fn(
@@ -125,7 +85,7 @@ class DriveTerm:
         frame_frequency: float,
         drive_base: complex,
         rwa: bool
-    ) -> Tuple[HamiltonianCoefficient, Union[HamiltonianCoefficient, None], float]:
+    ) -> Tuple[HamiltonianCoefficient, Union[HamiltonianCoefficient, None]]:
         r"""Generate the coefficients for X and Y drives.
 
         Args:
@@ -134,14 +94,8 @@ class DriveTerm:
             rwa: If True, returns the RWA coefficients.
 
         Returns:
-            A 3-tuple of X and Y coefficients and the maximum frequency appearing in the term.
+            X and Y coefficient functions.
         """
-        if isinstance(self.amplitude, PulseSequence):
-            return self.amplitude.generate_fn(frame_frequency, drive_base, rwa, initial_frequency=self.frequency)
-
-        if self.frequency is None:
-            raise RuntimeError('Drive frequency not set')
-
         amplitude = self.amplitude
         if isinstance(amplitude, str):
             # If this is actually a static expression, convert to complex
@@ -152,16 +106,14 @@ class DriveTerm:
 
         if rwa:
             fn_x, fn_y = self._generate_fn_rwa(amplitude, frame_frequency, drive_base)
-            max_freq = abs(self.frequency - frame_frequency)
         else:
             fn_x, fn_y = self._generate_fn_full(amplitude, frame_frequency, drive_base)
-            max_freq = self.frequency + frame_frequency
 
-        return fn_x, fn_y, max_freq
+        return fn_x, fn_y
 
     def _generate_fn_rwa(self, amplitude, frame_frequency, drive_base):
         detuning = self.frequency - frame_frequency
-        is_resonant = (abs(detuning) < REL_FREQUENCY_EPSILON * frame_frequency)
+        is_resonant = np.isclose(detuning, 0.)
 
         if isinstance(amplitude, (float, complex)):
             # static envelope
