@@ -1,17 +1,21 @@
-"""Pulse shape library."""
+r"""
+============================================
+Pulse shape library (:mod:`qudit_sim.pulse`)
+============================================
 
-from typing import List, Union, Optional
+.. currentmodule:: qudit_sim.pulse
+
+Classes in this module represent pulse envelopes. Subclasses of Pulse can be passed to
+``HamiltonianBuilder.add_drive`` either as the ``amplitude`` parameter or through the PulseSequence
+class (``sequence`` parameter).
+"""
+
+from typing import List, Union, Optional, Any
 import copy
-import logging
 from dataclasses import dataclass
 import numpy as np
 
-from .util import FrequencyScale, CallableCoefficient
-
-logger = logging.getLogger(__name__)
-
-s = FrequencyScale.Hz.time_value
-ns = FrequencyScale.GHz.time_value
+from .drive import CallableCoefficient
 
 class PulseSequence(list):
     """Pulse sequence.
@@ -21,12 +25,34 @@ class PulseSequence(list):
     list with a single additional function `generate_fn`.
     """
 
+    @property
+    def duration(self):
+        d = sum(inst.value for inst in self if isinstance(inst, Delay))
+        d += sum(inst.duration for inst in self if isinstance(inst, Pulse))
+        return d
+
+    @property
+    def frequency(self):
+        frequency = 0.
+        max_frequency = 0.
+        for inst in self:
+            if isinstance(inst, ShiftFrequency):
+                frequency += inst.value
+                max_frequency = max(max_frequency, frequency)
+            elif isinstance(inst, SetFrequency):
+                frequency = inst.value
+                max_frequency = max(max_frequency, frequency)
+
+        return max_frequency
+
+    def __str__(self):
+        return f'PulseSequence([{", ".join(str(inst) for inst in self)}])'
+
     def generate_fn(
         self,
         frame_frequency: float,
         drive_base: complex,
-        rwa: bool = True,
-        initial_frequency: Optional[float] = None
+        rwa: bool = False
     ) -> tuple:
         r"""Generate the X and Y drive coefficients and the maximum frequency appearing in the sequence.
 
@@ -42,26 +68,28 @@ class PulseSequence(list):
                 is SetFrequency.
 
         Returns:
-            A 3-tuple of X and Y coefficients and the maximum frequency appearing in the term.
+            X and Y coefficient functions.
         """
+        if len(self) == 0:
+            return 0., 0.
+
+        assert isinstance(self[0], SetFrequency), 'First instruction of a PulseSequence must be SetFrequency'
+
         funclist_x = []
         funclist_y = []
         timelist = []
 
-        frequency = initial_frequency
-        max_frequency = 0. if frequency is None else frequency
+        frequency = None
         phase_offset = 0.
         time = 0.
 
         for inst in self:
             if isinstance(inst, ShiftFrequency):
                 frequency += inst.value
-                max_frequency = max(max_frequency, frequency)
             elif isinstance(inst, ShiftPhase):
                 phase_offset += inst.value
             elif isinstance(inst, SetFrequency):
                 frequency = inst.value
-                max_frequency = max(max_frequency, frequency)
             elif isinstance(inst, SetPhase):
                 phase_offset = inst.value - frequency * time
             elif isinstance(inst, Delay):
@@ -79,20 +107,61 @@ class PulseSequence(list):
 
         timelist.append(time)
 
-        def make_fn(timelist, funclist):
-            def fn(t, args):
-                # piecewise determines the output dtype from tlist
-                t = np.asarray(t, dtype=np.float)
-                condlist = [((t >= start) & (t < end)) for start, end in zip(timelist[:-1], timelist[1:])]
-                return np.piecewise(t, condlist, funclist, args)
+        fn_x = _make_sequence_fn(timelist, funclist_x)
+        fn_y = _make_sequence_fn(timelist, funclist_y)
 
-            return fn
+        return fn_x, fn_y
 
-        fn_x = make_fn(timelist, funclist_x)
-        fn_y = make_fn(timelist, funclist_y)
+    def envelope(self, t: Union[float, np.ndarray], args: Any = None) -> np.ndarray:
+        """Return the envelope of the sequence as a function of time.
 
-        return fn_x, fn_y, max_frequency
+        This function is mostly for visualization purposes. Phase and frequency information is lost in the
+        returned array.
 
+        Args:
+            t: Time or array of time points.
+            args: Second argument to the pulse envelope functions.
+
+        Returns:
+            Pulse sequence envelope (complex) as a function of time.
+        """
+        funclist = []
+        timelist = []
+
+        time = 0.
+
+        for inst in self:
+            if isinstance(inst, Delay):
+                funclist.append(0.)
+                timelist.append(time)
+                time += inst.value
+            elif isinstance(inst, Pulse):
+                funclist.append(_make_shifted_fn(inst, time))
+                timelist.append(time)
+                time += inst.duration
+
+        timelist.append(time)
+
+        # piecewise determines the output dtype from tlist
+        tlist = np.asarray(t, dtype=np.complex128)
+        condlist = [((t >= start) & (t < end)) for start, end in zip(timelist[:-1], timelist[1:])]
+        return np.piecewise(tlist, condlist, funclist, args)
+
+
+def _make_sequence_fn(timelist, funclist):
+    def fn(t, args):
+        # piecewise determines the output dtype from tlist
+        tlist = np.asarray(t, dtype=np.float)
+        condlist = [((t >= start) & (t < end)) for start, end in zip(timelist[:-1], timelist[1:])]
+        return np.piecewise(tlist, condlist, funclist, args)
+
+    return fn
+
+def _make_shifted_fn(pulse, start_time):
+    def fn(t, args):
+        return pulse(t - start_time, args)
+
+    return fn
 
 class Pulse:
     """Base class for all pulse shapes.
@@ -105,7 +174,7 @@ class Pulse:
         assert duration > 0., 'Pulse duration must be positive'
         self.duration = duration
 
-    def __call__(self, t, args):
+    def __call__(self, t, args=None):
         raise NotImplementedError('Pulse is an ABC')
 
     def __mul__(self, c):
@@ -116,6 +185,9 @@ class Pulse:
         instance._scale(c)
         return instance
 
+    def __str__(self):
+        return f'Pulse(duration={self.duration})'
+
     def modulate(
         self,
         drive_base: complex,
@@ -146,7 +218,7 @@ class Pulse:
             else:
                 value = lambda envelope, phase: envelope.imag * np.cos(phase) - envelope.real * np.sin(phase)
 
-            def fun(t, args):
+            def fun(t, args=None):
                 envelope = drive_base * self.__call__(t - start_time, args)
                 phase = (frequency - frame_frequency) * t + phase_offset
                 return value(envelope, phase)
@@ -157,8 +229,8 @@ class Pulse:
             else:
                 frame_fn = np.sin
 
-            def fun(t, args):
-                double_envelope = 2. * drive_base * self.__call__(t - time, args)
+            def fun(t, args=None):
+                double_envelope = 2. * drive_base * self.__call__(t - start_time, args)
                 phase = frequency * t + phase_offset
                 prefactor = double_envelope.real * np.cos(phase) + double_envelope.imag * np.sin(phase)
                 return prefactor * frame_fn(frame_frequency * t)
@@ -187,8 +259,6 @@ class Gaussian(Pulse):
     ):
         super().__init__(duration)
 
-        assert sigma > 1. * ns, 'Gaussian sigma must be greater than 1 ns'
-
         self.sigma = sigma
 
         if center is None:
@@ -197,24 +267,32 @@ class Gaussian(Pulse):
             self.center = center
 
         if zero_ends:
-            assert np.abs(self.center - self.duration * 0.5) < 1. * ns, \
+            assert np.isclose(self.center, self.duration * 0.5), \
                 'Zero-ended Gaussian must be centered'
 
             x = self.duration * 0.5 / self.sigma
-            self.offset = np.exp(-np.square(x) * 0.5)
-            self.amp = np.asarray(amp / (1. - self.offset), dtype=np.complex128)
+            self._pedestal = np.exp(-np.square(x) * 0.5)
+            self._gaus_amp = np.asarray(amp / (1. - self._pedestal), dtype=np.complex128)
 
         else:
-            self.amp = np.asarray(amp, dtype=np.complex128)
-            self.offset = 0.
+            self._pedestal = 0.
+            self._gaus_amp = np.asarray(amp, dtype=np.complex128)
 
-    def __call__(self, t, args):
+    @property
+    def amp(self) -> complex:
+        return complex(self._gaus_amp * (1. - self._pedestal))
+
+    def __call__(self, t, args=None):
         x = (t - self.center) / self.sigma
-        return np.asarray(self.amp * (np.exp(-np.square(x) * 0.5) - self.offset),
+        return np.asarray(self._gaus_amp * (np.exp(-np.square(x) * 0.5) - self._pedestal),
                           dtype=np.complex128)
 
+    def __str__(self):
+        return (f'Gaussian(duration={self.duration}, amp={self.amp}, sigma={self.sigma}, center={self.center},'
+                f' zero_ends={self._pedestal != 0.})')
+
     def _scale(self, c):
-        self.amp *= c
+        self._gaus_amp *= c
 
 
 class GaussianSquare(Pulse):
@@ -233,38 +311,86 @@ class GaussianSquare(Pulse):
         amp: Union[float, complex],
         sigma: float,
         width: float,
-        zero_ends: bool = True
+        zero_ends: bool = True,
+        rise: bool = True,
+        fall: bool = True
     ):
+        if not (rise or fall):
+            raise ValueError("That's just a square pulse, dude")
+
         super().__init__(duration)
 
-        assert sigma > 1. * ns, 'Gaussian sigma must be greater than 1 ns'
         assert width < duration, 'GaussianSquare width must be less than duration'
 
-        self.t1 = (duration - width) / 2.
-
-        self.amp = amp
+        self.sigma = sigma
         self.width = width
 
-        self.gauss_left = Gaussian(duration=self.t1 * 2., amp=amp, sigma=sigma,
-                                   center=None, zero_ends=zero_ends)
-        self.gauss_right = Gaussian(duration=self.t1 * 2., amp=amp, sigma=sigma,
-                                    center=None, zero_ends=zero_ends)
+        self._condlist = list()
+        self._funclist = list()
 
-    def _right_tail(self, t, args):
-        return self.gauss_right(t - self.width, args)
+        ramp_duration = duration - width
+        if rise and fall:
+            gauss_duration = ramp_duration
+        else:
+            gauss_duration = ramp_duration * 2.
 
-    def __call__(self, t, args):
-        # piecewise determines the output dtype from the first argument
+        if rise:
+            self.t_plateau = gauss_duration / 2.
+            self.gauss_rise = Gaussian(duration=gauss_duration, amp=amp, sigma=sigma,
+                                       center=None, zero_ends=zero_ends)
+
+            self._condlist.append(lambda t: t < self.t_plateau)
+            self._funclist.append(self.gauss_rise)
+        else:
+            self.t_plateau = 0.
+            self.gauss_rise = None
+
+        if fall:
+            self.gauss_fall = Gaussian(duration=gauss_duration, amp=amp, sigma=sigma,
+                                       center=None, zero_ends=zero_ends)
+
+            self._condlist.append(lambda t: t >= self.t_plateau + self.width)
+            self._funclist.append(self._fall_tail)
+        else:
+            self.gauss_fall = None
+
+        self._funclist.append(complex(amp))
+
+    @property
+    def amp(self) -> complex:
+        return self._funclist[-1]
+
+    def _fall_tail(self, t, args):
+        gauss_t0 = self.t_plateau + self.width - self.gauss_fall.duration / 2.
+        return self.gauss_fall(t - gauss_t0, args)
+
+    def _make_condlist(self, t):
+        return list(cond(t) for cond in self._condlist)
+
+    def __call__(self, t, args=None):
+        # np.piecewise determines the output dtype from the first argument
         tlist = np.asarray(t, dtype=np.complex128)
-        return np.piecewise(tlist,
-            [t < self.t1, t >= self.t1 + self.width],
-            [self.gauss_left, self._right_tail, self.amp],
-            args)
+        return np.piecewise(tlist, self._make_condlist(t), self._funclist, args)
+
+    def __str__(self):
+        rise = self.gauss_rise is not None
+        fall = self.gauss_fall is not None
+
+        if rise:
+            zero_ends = self.gauss_rise._pedestal != 0.
+        else:
+            zero_ends = self.gauss_fall._pedestal != 0.
+
+        return (f'GaussianSquare(duration={self.duration}, amp={self.amp}, sigma={self.sigma}, width={self.width}, '
+                f' zero_ends={zero_ends}, rise={rise}, fall={fall})')
 
     def _scale(self, c):
-        self.amp *= c
-        self.gauss_left._scale(c)
-        self.gauss_right._scale(c)
+        self._funclist[-1] *= c
+
+        if self.gauss_rise:
+            self.gauss_rise._scale(c)
+        if self.gauss_fall:
+            self.gauss_fall._scale(c)
 
 
 class Drag(Gaussian):
@@ -298,12 +424,42 @@ class Drag(Gaussian):
 
         self.beta = beta
 
-    def __call__(self, t, args):
+    def __call__(self, t, args=None):
         gauss = super().__call__(t, args)
         dgauss = -(t - self.center) / np.square(self.sigma) * gauss
 
         return gauss + 1.j * self.beta * dgauss
 
+    def __str__(self):
+        return (f'Drag(duration={self.duration}, amp={self.amp}, sigma={self.sigma}, beta={self.beta}, '
+                f'center={self.center}, zero_ends={self._pedestal != 0.})')
+
+
+class Square(Pulse):
+    """Square (constant) pulse.
+
+    Args:
+        duration: Pulse duration.
+        amp: Pulse height relative to the drive base amplitude.
+    """
+    def __init__(
+        self,
+        duration: float,
+        amp: Union[float, complex]
+    ):
+        super().__init__(duration)
+
+        self.amp = complex(amp)
+
+    def __call__(self, t, args=None):
+        return np.full_like(t, self.amp)
+
+    def __str__(self):
+        return f'Square(duration={self.duration}, amp={self.amp})'
+
+    def _scale(self, c):
+        self.amp *= c
+
 
 @dataclass(frozen=True)
 class ShiftFrequency:
